module hybrid.GlGuiRenderer;


private {
	import hybrid.GuiRenderer;
	import hybrid.Utils;
	import hybrid.Font;
	import maths.Vec;
	
	import hybrid.IconCache;
	import utils.Memory;
	import hybrid.Texture;
	
	import derelict.opengl.gl;
	import derelict.opengl.glu;
	import derelict.opengl.extension.ext.blend_color;
}



private class GlTexture : Texture {
	int id;
}



class GlGuiRenderer : GuiRenderer, TextureMngr {
	
	// TextureMngr
	Texture createTexture(vec2i size, vec4 defColor) {
		ubyte[] data;
		data.alloc(size.x * size.y * 4);

		uint bitspp = 32;
		uint bytespp = bitspp / 8;
		
		ubyte f2ub(float f) {
			if (f < 0) return 0;
			if (f > 1) return 255;
			return cast(ubyte)rndtol(f * 255);
		}

		for (uint i = 0; i < size.x; ++i) {
			for (uint j = 0; j < size.y; ++j) {			
				for (uint c = 0; c < bytespp; ++c) {
					data[(size.x * j + i) * bytespp + c] = f2ub(defColor.cell[c]);
				}
			}
		}
		
		GlTexture tex = new GlTexture;
		glGenTextures(1, cast(uint*)&tex.id);
		
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, tex.id);
				
		const uint level = 0;
		const uint border = 0;
		GLenum format	= GL_RGBA;
		
		glTexImage2D(GL_TEXTURE_2D, level, bytespp, size.x, size.y, border, format, GL_UNSIGNED_BYTE, data.ptr);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,	GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,	GL_LINEAR);

		glDisable(GL_TEXTURE_2D);
		data.free();
		
		return tex;
	}
	
	
	// TextureMngr
	void updateTexture(Texture tex_, vec2i origin, vec2i size, ubyte[] data) {
		GlTexture tex = cast(GlTexture)(tex_);
		assert (tex !is null);
		
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, tex.id);
		
		const int level = 0;
		glTexSubImage2D(GL_TEXTURE_2D, level, origin.x, origin.y, size.x, size.y, GL_RGBA, GL_UNSIGNED_BYTE, data.ptr);
		
		glDisable(GL_TEXTURE_2D);
	}


	override void initialize() {
		DerelictGL.load();
		DerelictGLU.load();
		
		iconCache.texMngr = this;
		FontMngr.fontRenderer = this;
		
		DerelictGL.loadExtensions();
		if (glBlendColorSupported) {
			printf(`Using GL_EXT_BLEND_COLOR`\n);
		} else {
			printf(`NOT Using GL_EXT_BLEND_COLOR`\n);
		}
		
		setupGL();
	}
	
	
	private {
		// horizontal and vertical screen resolution
		const int   xResolution     = 800;
		const int   yResolution     = 600;
		
		// number of bits per pixel used for display. 24 => true color
		const int   bitsPerPixel    = 24;
		
				
		/**
		  Setup some basic OpenGL parameters
		*/
		void setupGL() {
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			gluOrtho2D(0, xResolution, yResolution, 0);
			glDisable(GL_DEPTH_TEST);
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			
			glEnable(GL_COLOR_MATERIAL);

			// alloc texture space for whitespace
			{
				const int whiteSize = 3;
				whiteSpace.tex = iconCache.get(
						vec2i(whiteSize, whiteSize),
						whiteSpace.bl,
						whiteSpace.tr,
						whiteSpace.blCoords,
						whiteSpace.trCoords
					);

				enableTexturing(whiteSpace.tex);

				ubyte[whiteSize*whiteSize*4] white = 255;
				iconCache.updateTexture(whiteSpace.tex, whiteSpace.bl, vec2i(whiteSize, whiteSize), white);
				
				disableTexturing();
			}
				
			lineWidth(1);
		}
	}


	// be nice and release all resources
	protected void cleanup() {
		DerelictGLU.unload();
		DerelictGL.unload();
	}
	
	
	protected final bool glBlendColorSupported() {
		return glBlendColorEXT !is null;
	}
	
	
	void addQuad(inout Quad q) {
		quads_.append(q, &numQuads);
	}
	
	
	Quad[] quads() {
		return quads_[0..numQuads];
	}
	

	override void clearQuads() {
		numQuads = 0;
	}
	
	
	struct IntBuffer {
		int[]	data_;
		uint	len;
		
		int[]	opCall() {
			return data_[0..len];
		}
		
		void reset() {
			len = 0;
		}
		
		void opCatAssign(int i) {
			data_.append(i, &len);
		}
		
		int length() { return len; }
	}


	protected {
		IntBuffer	curQuadsToDraw_;
		IntBuffer	subpixelSecondPass;
		IntBuffer	nextQuadsToDraw_;
		IntBuffer	allQuadsToDraw;
	}
	

	// instead of using 'synchronized' here, the IntBuffers could be put into thread-local storage
	override synchronized final void drawAll() {
		curQuadsToDraw_.reset();
		subpixelSecondPass.reset();
		nextQuadsToDraw_.reset();
		allQuadsToDraw.reset();
		
		IntBuffer* quadsToDraw	= &curQuadsToDraw_;
		IntBuffer* nextQuads		= &nextQuadsToDraw_;
		
		void swapQuadBuffers() {
			auto t = quadsToDraw;
			quadsToDraw = nextQuads;
			nextQuads = t;
		}
		
		for (int i = 0; i < quads.length; ++i) {
			allQuadsToDraw ~= i;
		}
		
		Texture lastTex = null;
		BlendingMode blending = BlendingMode.None;
		
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_BLEND);
				
		Rect curClip;
		
		
		void setViewport() {
			if (curClip == Rect.init) {
				curClip.min = vec2i(0, 0);
				curClip.max = vec2i(xResolution, yResolution);
			}
			int w = curClip.max.x - curClip.min.x;
			int h = curClip.max.y - curClip.min.y;
			
			glViewport(curClip.min.x, yResolution - curClip.min.y - h, w, h);
			
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			gluOrtho2D(curClip.min.x, curClip.max.x, curClip.max.y, curClip.min.y);
			glMatrixMode(GL_MODELVIEW);
		}
		

		void drawQuadsWorker(int[] quadList, void delegate(int, inout Quad) dg) {
			foreach (quadi; quadList) {
				auto quad = &quads[quadi];
				if (quad.tex !is lastTex || quad.clipRect != curClip) {
					glEnd();
					if (quad.clipRect != curClip) {
						if (quad.clipRect == Rect.init && curClip.min == vec2i.zero && curClip.max == vec2i(xResolution, yResolution)) {
							// skip the viewport stuff - the clip rect covers the entire viewport
						} else {
							curClip = quad.clipRect;
							setViewport();
						}
					}
					if (quad.tex !is lastTex) {
						auto tex = cast(GlTexture)quad.tex;
						assert (tex !is null);
						glBindTexture(GL_TEXTURE_2D, tex.id);
					}
					glBegin(GL_QUADS);
				}
				
				dg(quadi, *quad);
			}
		}

		
		void drawSecondSubpixelPass() {
			glEnd();
			
			scope(success) {
				glBegin(GL_QUADS);
				subpixelSecondPass.reset();
			}

			glBlendFunc(GL_SRC_ALPHA, GL_ONE);
			drawQuadsWorker(subpixelSecondPass(), (int quadi, inout Quad quad) {
				glColor4fv(&quad.col.x);
				drawQuad(quad);
			});
		}
		

		Rect	subpixQuadSum;
		vec4	prevConstColor = vec4(-1, 0, 0, 0);	// invalid
		
		bool setBlending(int quadIdx, inout Quad q) {
			if (!glBlendColorSupported) {
				if (BlendingMode.Subpixel == q.blending) {
					if (BlendingMode.Subpixel == blending) {	// next subpixel quad
						Rect qrect = q.calcRect;
						
						--qrect.max.x;		// iirc, opengl doesnt render the box inclusively
					
						if (subpixQuadSum.intersect(qrect)) {		// collision, schedule for later.
							*nextQuads ~= quadIdx;
							return false;
						} else {		// no collision, just render it.
							subpixelSecondPass ~= quadIdx;
							
							/*
								The precise way would be to use a sum here, but it results in lots of incorrect matches, as the box enlarges by too much
								when entering a new line.
								This is an approximation. It assumes that if the current glyph doesnt intersect with the previous one, the next will not
								intersect with the previous either. Errors will occur if two lines are not vertically spaced apart, but these cases
								are pathological anyway.
							*/
							subpixQuadSum = qrect;     // the precise method: Rect.sum(qrect, subpixQuadSum);
							return true; // render normally
						}
					} else {		// first subpixel quad
						subpixQuadSum = q.calcRect;
						subpixelSecondPass ~= quadIdx;
					}
				}
				else if (BlendingMode.Subpixel == blending) {		// just ended drawing subpixel blended quads
					drawSecondSubpixelPass();

					while (nextQuads.length > 0) {
						swapQuadBuffers();
						blending = BlendingMode.None;
						
						drawQuadsWorker((*quadsToDraw)(), (int quadi, inout Quad quad) {
							if (setBlending(quadi, quad)) {
								drawQuad(quad);
							}
						});
						quadsToDraw.reset();
						drawSecondSubpixelPass();
					}
					swapQuadBuffers();
				}
			}
			
			switch (q.blending) {
				case BlendingMode.Alpha:
					glColor4fv(&q.col.x);
					break;
				case BlendingMode.Subpixel:
					glColor3f(q.col.a, q.col.a, q.col.a);
					if (glBlendColorSupported) {
						vec4 constCol = vec4(q.col.a*q.col.r, q.col.a*q.col.g, q.col.a*q.col.b, 1);
						if (prevConstColor != constCol) {
							glEnd();
							glBlendColorEXT(constCol.r, constCol.g, constCol.b, constCol.a);
							prevConstColor = constCol;
							glBegin(GL_QUADS);
						}
					}
					break;
				case BlendingMode.None:
					break;
			}
			
			if (blending == q.blending) return true;
			
			glEnd();
			switch (blending = q.blending) {
				case BlendingMode.Alpha:
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					break;
				case BlendingMode.Subpixel:
					if(glBlendColorSupported) glBlendFunc(GL_CONSTANT_COLOR_EXT, GL_ONE_MINUS_SRC_COLOR);
					else glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
					glEnable(GL_BLEND);
					break;
				case BlendingMode.None:
					glDisable(GL_BLEND);
					break;
			}
			glBegin(GL_QUADS);
			
			return true;
		}
		
		
		void drawQuads() {
			setViewport();

			glBegin(GL_QUADS);
			
			drawQuadsWorker(allQuadsToDraw(), (int quadi, inout Quad quad) {
				if (setBlending(quadi, quad)) {
					drawQuad(quad);
				}
			});
			quadsToDraw.reset();
			
			Quad q;
			q.blending = BlendingMode.None;
			setBlending(0, q);
			
			glEnd();
			curClip = Rect.init;
			setViewport();
		}
		
		drawQuads();
		clearQuads();
	}
	

	private final void drawQuad(inout Quad q) {
		for (int i = 0; i < 4; ++i) {
			glTexCoord2fv(&q.tc[i].x);
			glVertex2fv(&q.pt[i].x);
		}
	}

	
	private {
		Quad[]				quads_;
		uint					numQuads;
	}
}
