module hybrid.GuiRenderer;

private {
	import hybrid.Utils;
	import hybrid.Font;
	import hybrid.FontRenderer;
	import hybrid.IconCache;
	import hybrid.Texture;
	
	import maths.Vec;
	import utils.Memory;
	import utils.StructClass;
}

alias hybrid.FontRenderer.BlendingMode BlendingMode;



private struct LineBrush {
	int	w;
	int	blur;
	mixin simpleStructCtor;
}


/**
	Abstract quad-based renderer. Renders everything as textured quads, even lines.
*/
abstract class GuiRenderer : FontRenderer {
	abstract void drawAll();
	abstract void clearQuads();
	
	
	void initialize() {
	}


	void setOffset(vec2 off) {
		m_offset = off;
	}
	
	
	void offset(vec2 off) {
		m_offset += off;
	}
	
	
	vec2 getOffset() {
		return m_offset;
	}
	
	
	void polygon(vec2[] points ...) {
		void point(vec2 pt) {
			pt += m_offset;
			this.absoluteQuadPoint(pt);
		}
		
		foreach (i, vec2 pt; points[1..$]) {
			point(points[0]);
			point(points[i]);
			point(pt);
			point(pt);
		}
	}
	
	
	void rect(Rect r) {
		polygon(vec2.convert(r.min), vec2(r.min.x, r.max.y), vec2.convert(r.max), vec2(r.max.x, r.min.y));
	}


	void polyLine(vec2[] points ...) {
		prevLinePt = points[$-1] + m_offset;
		foreach (vec2 pt; points) {
			pt += m_offset;
			drawLine(prevLinePt, pt);
			prevLinePt = pt;
		}

		disableTexturing();
	}
	
	
	void drawLine(vec2 prev, vec2 cur) {
		enableTexturing(curLineBrush.tex);
		float len = (cur-prev).len;
		vec2 off = (cur-prev).normalized * 0.25;
		cur += off;
		prev += off;
		
		int width = curLineSetting.w;
		vec2 leftoff = (cur - prev).rotatedHalfPi.normalized * width*0.5;// + vec2(0.5, 0.5);
		vec2 rightoff = (cur - prev).rotatedHalfPi.normalized * width*0.5;// - vec2(0.5, 0.5);
		
		void point(vec2 tc, vec2 pt) {
			this.setTC(tc);
			this.absoluteQuadPoint(pt);
		}
		
		auto pen = curLineBrush;
		point(vec2(pen.blCoords.x, pen.trCoords.y), cur+rightoff);
		point(vec2(pen.trCoords.x, pen.trCoords.y), cur-leftoff);
		point(vec2(pen.trCoords.x, pen.blCoords.y), prev-leftoff);
		point(vec2(pen.blCoords.x, pen.blCoords.y), prev+rightoff);
	}
	
	
	void line(vec2[] points ...) {
		foreach (i, vec2 pt; points) {
			pt += m_offset;
			if (i != 0) drawLine(prevLinePt, pt);
			prevLinePt = pt;
		}
		
		disableTexturing();
	}


	void rectLine(Rect r) {
		polyLine(vec2.convert(r.min), vec2(r.min.x, r.max.y), vec2.convert(r.max), vec2(r.max.x, r.min.y));
	}
	
	
	void color(vec4 col) {
		curColor = col;
	}
	

	void color(vec3 rgb, float a) {
		curColor = vec4(rgb.r, rgb.g, rgb.b, a);
	}
	
	
	void blendingMode(BlendingMode m) {
		curBlendingMode = m;
	}
	
	
	void disableTexturing() {
		enableTexturing(whiteSpace.tex);
		vec2 tc = (whiteSpace.blCoords + whiteSpace.trCoords) / 2;
		setTC(tc);
	}
	

	void enableTexturing(Texture tex) {
		curTexture = tex;
	}

	
	void lineWidth(int w, int blur = 0) {
		if (w < 1) w = 1;
		if (blur < 0) blur = 0;
		auto brush = LineBrush(w, blur);
		curLineBrush = getLineStroke(brush);
		curLineSetting = brush;
	}
	

	void print(Font f, vec2 pos, char[] text) {
		f.print(vec2i.convert(pos + m_offset), curColor, text);
	}
	

	void print(Font f, vec2 pos, dchar[] text) {
		f.print(vec2i.convert(pos + m_offset), curColor, text);
	}
	
	
	LoadedIcon createLineStroke(int width, int blur) {
		LoadedIcon res;
		
		vec2i size = vec2i(width+blur*2, 3);
		res.tex = iconCache.get(size, res.bl, res.tr, res.blCoords, res.trCoords);
		res.blCoords += iconCache.texelSize;
		res.trCoords -= iconCache.texelSize;
		assert (res.tex !is null);
		
		FontMngr.fontRenderer.enableTexturing(res.tex);
		
		ubyte[] pen;
		pen.alloc(size.x * size.y * 4);
		
		pen[] = 0;
		{
			for (int y = 0; y < size.y; ++y) {
				//for (int x = 1; x < size.x-1; ++x) {
				for (int x = 0; x < size.x; ++x) {
					int off = (y*size.x+x)*4;
					int val = 255;
					if (blur > 0) {
						if (x < blur) val = 255 * x / blur;
						int xe = size.x - x - 1;
						if (xe < blur) val = 255 * xe / blur;
					}
					pen[off..off+4] = val;
				}
			}
		}

		iconCache.updateTexture(res.tex, res.bl, size, pen);
		pen.free();
		return res;
	}
	
	
	LoadedIcon getLineStroke(LineBrush b) {
		LoadedIcon* res = b in lineBrushes;
		if (res) return *res;
		return lineBrushes[b] = createLineStroke(b.w/2+1, b.blur);
	}	
	
	
	void setTC(vec2 tc) {
		curTC = tc;
	}
	
	
	void quadPoint(vec2 pt) {
		absoluteQuadPoint(pt+m_offset);
	}
	
	
	void absoluteQuadPoint(vec2 pt) {
		curQuad.tc[curPoint] = curTC;
		curQuad.pt[curPoint++] = pt;
		if (4 == curPoint) {
			curQuad.tex = curTexture;
			curQuad.col = curColor;
			curQuad.clipRect = curClipRect;
			curQuad.blending = curBlendingMode;
			addQuad(curQuad);
			curPoint = 0;
		}
	}
	
	
	void clip(Rect rect) {
		curClipRect = rect;
	}
	
	
	void resetClipping() {
		curClipRect = Rect.init;
	}
	
	
	abstract protected void addQuad(inout Quad q);



	protected {
		private struct Quad {
			vec2[4]				pt;
			vec2[4]				tc;
			vec4					col;
			Texture				tex;
			Rect					clipRect;
			BlendingMode	blending;
			
			// returns a rect enclosing this quad
			Rect	calcRect() {
				int maxX = int.min;
				int minX = int.max;
				int maxY = int.min;
				int minY = int.max;
				foreach (p; pt) {
					int l = cast(int)p.x;
					int r = cast(int)(p.x+0.99);
					int b = cast(int)p.y;
					int t = cast(int)(p.y+0.99);
					if (l < minX) minX = l;
					if (t < minY) minY = t;
					if (r > maxX) maxX = r;
					if (b > maxY) maxY = b;
				}
				Rect r;
				r.min = vec2i(minX, minY);
				r.max = vec2i(maxX, maxY);
				return r;
			}
		}
		

		private struct LoadedIcon {
			Texture	tex;
			vec2i	size;
			vec2i 	bl;
			vec2i	tr;
			vec2		blCoords;
			vec2		trCoords;
		}
	}


	protected {
		vec2					m_offset = {x:0, y:0};
		
		LoadedIcon					whiteSpace;
		LoadedIcon[LineBrush]	lineBrushes;
		LoadedIcon					curLineBrush;
		LineBrush						curLineSetting;

		vec2					prevLinePt;
		Texture				curTexture;
		vec4					curColor;
		vec2					curTC;
		Rect					curClipRect;
		BlendingMode	curBlendingMode = BlendingMode.Alpha;
		
		Quad				curQuad;
		int					curPoint;
	}
}
